unit Unit1;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls;

type

  TForm1 = class(TForm)
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    Button4: TButton;
    Button5: TButton;
    Button6: TButton;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);
    procedure Button6Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;



  IVeiculo = interface  (iinterface)
  ['{1D2F0F1D-60CC-47BE-AD22-4DA59D87C7CA}']
    procedure DeslocarSe;
    procedure Parar;
  end;

  IVeiculoTerrestre = interface (IVeiculo)
  ['{CD718BB9-FE3A-4C19-8CE2-01EBB1843B96}']
    procedure DeslocarSe;
    procedure Parar;
    procedure Rodar;
  end;

  ICarro = interface (IVeiculoTerrestre)
  ['{3B167780-8EA0-4F59-9DBF-A9F68F54F595}']
    procedure DeslocarSe;
    procedure Parar;
    procedure Rodar;
  end;

  //TCarro = class(TInterfacedObject,  ICarro)  //nessa forma voc pode atribuir a criao de um TCarro a uma varivel ICarro, IVeiculoTerrestre e IVeiculo, mas no pode fazer converses do tipo AS entre eles
  TCarro = class(TInterfacedObject,  ICarro, IVeiculoTerrestre, IVeiculo) //esta forma permite qualquer tipo de atribuio e converso.
  private
    FPlaca: TStringList;
    function GetPlaca: string;
    procedure SetPlaca(value: string);
  public
    constructor Create;
    destructor Destroy; override;

    procedure DeslocarSe;
    procedure Parar;
    procedure Rodar;
    property Placa: string read GetPlaca write SetPlaca;
  end;


var
  Form1: TForm1;


implementation

{$R *.dfm}

{ TCarro }

constructor TCarro.Create;
begin
  FPlaca := TStringList.Create;
end;

procedure TCarro.DeslocarSe;
begin
  ShowMessage(Placa + ' Deslocando-se. ' + self.RefCount.ToString() + ' referncias');
end;

destructor TCarro.Destroy;
begin
  FPlaca.Free;
  inherited;
end;

function TCarro.GetPlaca: string;
begin
  Result := FPlaca.Text;
end;

procedure TCarro.Parar;
begin
  ShowMessage(Placa + ' Parando.' + self.RefCount.ToString() + ' referncias');
end;

procedure TCarro.Rodar;
begin
  DeslocarSe;
end;

procedure TCarro.SetPlaca(value: string);
begin
  FPlaca.Text := value;
end;

procedure TForm1.Button1Click(Sender: TObject);
var C: TCarro;
begin

  C := TCarro.Create; //tenho um objeto, mas C no conta referencias
  C.Placa := 'ABC1234';
  (C as ICarro).Rodar; //converte e roda na mesma linha, mas zera o contador de referencias na proxima linha
  C.Parar;  //isso d erro
  C.Rodar;
  C.Parar;

end;

procedure TForm1.Button2Click(Sender: TObject);
var C: TCarro;
  I: IVeiculoTerrestre;
begin

  C := TCarro.Create; //tenho um objeto, mas C no conta referencias
  C.Placa := 'ABC1234';
  I := (C as ICarro); //Converso joga  para I o controle da contagem de referencias
  I.Rodar; //uso normal
  I.Parar; //uso normal
  I.Rodar; //uso normal
  I.Parar; //uso normal

  C.Rodar; //Funciona normal, mas no se deve misturar objetos comuns com interfaces, por isso para de usar C uma vez convertido para I
  C.Parar; //Funciona normal, mas no se deve misturar objetos comuns com interfaces, por isso para de usar C uma vez convertido para I

end;

procedure TForm1.Button3Click(Sender: TObject);
var C: TCarro;
  I: IVeiculo;
begin

  C := TCarro.Create; //tenho um objeto, mas C no conta referencias
  C.Placa := 'ABC1234';
  I := (C as ICarro); //Converso joga  para I o controle da contagem de referencias

  I.DeslocarSe;
  I.Parar;

  //I agora pode ser convertido para qualquer uma das interfaces ancestrais de ICarro sem perder o controle de referncia
  (I as ICarro).Rodar;
  (I as ICarro).Parar;
  (I as IVeiculoTerrestre).Rodar;
  (I as IVeiculoTerrestre).Parar;
  (I as IVeiculo).DeslocarSe;
  (I as IVeiculo).Parar;

  C.Rodar; //Funciona normal, mas no se deve misturar objetos comuns com interfaces, por isso para de usar C uma vez convertido para I
  C.Parar; //Funciona normal, mas no se deve misturar objetos comuns com interfaces, por isso para de usar C uma vez convertido para I

end;

procedure TForm1.Button4Click(Sender: TObject);
var C: TCarro;
  iv1, iv2: IVeiculo;
  ivt1, ivt2: IVeiculoTerrestre;
  ic: ICarro;
begin

  C := TCarro.Create;
  C.Placa := 'ABC1234';
  ic  := (C as ICarro);

  ivt1 := (C as IVeiculoTerrestre);
  iv1  :=  (C as IVeiculo);

  ivt2 := (ic as IVeiculoTerrestre);
  iv2  :=  (ic as IVeiculo);

  ic.DeslocarSe;
  ic.Parar;
  ivt1.Rodar;
  ivt1.Parar;
  iv1.DeslocarSe;
  iv1.Parar;

  ivt2.Rodar;
  ivt2.Parar;
  iv2.DeslocarSe;
  iv2.Parar;

end;

procedure TForm1.Button5Click(Sender: TObject);
var C: TCarro;
  iv1, iv2: IVeiculo;
  ivt1, ivt2: IVeiculoTerrestre;
  ic: ICarro;
begin

  C := TCarro.Create;
  C.Placa := 'ABC1234';
  ic  := (C as ICarro);

  ic.DeslocarSe;
  ic.Parar;

  if Supports(c, IVeiculoTerrestre, ivt1) then
  begin
    ivt1.Rodar;
    ivt1.Parar;
  end;

  if Supports(c, IVeiculo, iv1) then
  begin
    iv1.DeslocarSe;
    iv1.Parar;
  end;

  if Supports(ic, IVeiculoTerrestre, ivt2) then
  begin
    ivt2.Rodar;
    ivt2.Parar;
  end;

  if Supports(ic, IVeiculo, iv2) then
  begin
    iv2.DeslocarSe;
    iv2.Parar;
  end;

end;

procedure TForm1.Button6Click(Sender: TObject);
var C: TCarro;
  iv1, iv2: IVeiculo;
  ivt1, ivt2: IVeiculoTerrestre;
  ic: ICarro;
begin

  C := TCarro.Create;
  C.Placa := 'ABC1234';
  ic  := ICarro(C);

  ivt1 := IVeiculoTerrestre(C);
  iv1  := IVeiculo(C);

  ivt2 :=  IVeiculoTerrestre(ic);
  iv2  :=  IVeiculo(ic);

  ic.DeslocarSe;
  ic.Parar;
  ivt1.Rodar;
  ivt1.Parar;
  iv1.DeslocarSe;
  iv1.Parar;

  ivt2.Rodar;
  ivt2.Parar;
  iv2.DeslocarSe;
  iv2.Parar;
end;

end.
